@@ -19,6 +19,78 @@ class Test_Push_Forward(TestCase):
         for n in range(3):
             for m in range(4):
                 assert_array_almost_equal(z[0,0].data, A[0,0].data)
+
+    def test_sum(self):
+        # check sum
+        D,P,N,M = 2,3,4,5
+        x = UTPM(numpy.arange(D*P*N*M).reshape((D,P,N,M)))
+        
+        y0 = numpy.sum(x).data
+        y1 = numpy.sum(x,axis=0).data
+        y2 = numpy.sum(x,axis=1).data
+        
+        for d in range(D):
+            for p in range(P):
+                assert_array_almost_equal(numpy.sum(x.data[d,p]), y0[d,p])
+                assert_array_almost_equal(numpy.sum(x.data[d,p], axis=0), y1[d,p])
+                assert_array_almost_equal(numpy.sum(x.data[d,p], axis=1), y2[d,p])
+        
+
+    def test_broadcasting_sub(self):
+        #check 1
+        x1 = numpy.array([1.,2.,3.])
+        y1 = UTPM([[5],[7]])
+        
+        z11 = x1 - y1
+        z12 = -(y1 - x1)
+        
+        z_data = numpy.array([[[-4,-3,-2]],[[-7,-7,-7]]])
+        assert_array_almost_equal(z_data, z11.data)
+        assert_array_almost_equal(z_data, z12.data)
+        
+        
+    def test_broadcasting_setitem(self):
+        x = UTPM(numpy.arange(2*1*3*4).reshape((2,1,3,4)))
+        y = UTPM(numpy.arange(2*1).reshape((2,1)))
+        x[0, :] = y
+        x[:, 1] = y
+        
+        assert_array_almost_equal(0, x.data[0,:,0,:])
+        assert_array_almost_equal(1, x.data[1,:,0,:])
+        assert_array_almost_equal(0, x.data[0,:,:,1])
+        assert_array_almost_equal(1, x.data[1,:,:,1])
+        
+        x[0, :] = 3.
+        x[:, 1] = 3.
+        
+        assert_array_almost_equal(3, x.data[0,:,0,:])
+        assert_array_almost_equal(0, x.data[1,:,0,:])
+        assert_array_almost_equal(3, x.data[0,:,:,1])
+        assert_array_almost_equal(0, x.data[1,:,:,1])
+
+    
+
+                
+    def test_symvec_vecsym(self):
+        (D,P,N) = 2,1,5
+        A = UTPM(numpy.random.rand(*(D,P,N,N)))
+        A = UTPM.dot(A.T,A)
+        v = UTPM.symvec(A)
+        B = UTPM.vecsym(v)
+        
+        assert_array_almost_equal(A.data, B.data)
+        
+        
+    def test_symvec_vecsym_pullback(self):
+        (D,P,N) = 2,1,6
+        v = UTPM(numpy.random.rand(*(D,P,N)))
+        A = UTPM.vecsym(v)
+        w = UTPM.symvec(A)
+        wbar = UTPM(numpy.random.rand(*(D,P,N)))
+        Abar = UTPM.pb_symvec(wbar, A, 'F', w)
+        vbar = UTPM.pb_vecsym(Abar, v, A)
+        
+        assert_array_almost_equal(wbar.data, vbar.data)
     
     
     def test_UTPM_in_a_stupid_way(self):
@@ -241,14 +313,94 @@ class Test_Push_Forward(TestCase):
         Z = X*X*X
         assert_array_almost_equal(Y.data, Z.data)
         
-    def test_sin_cos(self):
+        
+    def test_pow_pullback(self):
+        D,P,N = 4,2,2
+        x = UTPM(numpy.random.rand(D,P,N))
+        
+        r = 2
+        y = x**r
+        ybar = UTPM(numpy.random.rand(D,P,N))
+        xbar = UTPM.pb___pow__(ybar, x, r, y)
+        
+        assert_array_almost_equal((r * ybar * x**(r-1)).data, xbar.data)
+        
+        
+        r = 3.1
+        y = x**r
+        ybar = UTPM(numpy.random.rand(D,P,N))
+        xbar = UTPM.pb___pow__(ybar, x, r, y)
+        
+        assert_array_almost_equal((r * ybar * x**(r-1)).data, xbar.data)
+        
+    def test_sincos(self):
         D,P,M,N = 4,3,2,1
         X = UTPM(numpy.random.rand(D,P,M,N))
         Z = UTPM.sin(X)**2 + UTPM.cos(X)**2
         # print Z
         assert_array_almost_equal(Z.data[0], numpy.ones((P,M,N)))
         assert_array_almost_equal(Z.data[1], numpy.zeros((P,M,N)))
+
+    def test_tansec(self):
+        D,P,M,N = 4,3,2,1
         
+        x = UTPM(numpy.random.random((D,P,M,N)))
+        y1 = UTPM.tan(x)
+        y2 = UTPM.sin(x)/UTPM.cos(x)
+        assert_array_almost_equal(y2.data, y1.data)
+        
+    def test_arcsin(self):
+        D,P,N,M = 5,3,4,5
+        x = UTPM(numpy.random.random((D,P,M,N)))
+        
+        y = UTPM.arcsin(x)
+        x2 = UTPM.sin(y)
+        y2 = UTPM.arcsin(x2)
+        
+        assert_array_almost_equal(x.data, x2.data)
+        assert_array_almost_equal(y.data, y2.data)
+        
+    def test_arccos(self):
+        D,P,N,M = 5,3,4,5
+        x = UTPM(numpy.random.random((D,P,M,N)))
+        
+        y = UTPM.arccos(x)
+        x2 = UTPM.cos(y)
+        y2 = UTPM.arccos(x2)
+        
+        assert_array_almost_equal(x.data, x2.data)
+        assert_array_almost_equal(y.data, y2.data)
+        
+        
+        
+    def test_arctan(self):
+        D,P,N,M = 5,3,4,5
+        x = UTPM(numpy.random.random((D,P,M,N)))
+        y  = UTPM.tan(x)
+        x2 = UTPM.arctan(y)
+        y2  = UTPM.tan(x2)
+        assert_array_almost_equal(x.data, x2.data)
+        assert_array_almost_equal(y.data, y2.data)
+        
+        
+
+    def test_sinhcosh(self):
+        D,P,N,M = 5,3,4,5
+        x = UTPM(numpy.random.random((D,P,M,N)))
+        y = UTPM.sinh(x)
+        z = UTPM.cosh(x)
+        assert_array_almost_equal(0., (z**2 - y**2 - 1.).data)
+        
+    def test_tanh(self):
+        D,P,N,M = 5,3,4,5
+        x = UTPM(numpy.random.random((D,P,M,N)))
+        
+        s = UTPM.sinh(x)
+        c = UTPM.cosh(x)
+        t = UTPM.tanh(x)
+        assert_array_almost_equal(t.data, (s/c).data)
+        
+
      
     def test_abs(self):
         D,P,N = 4,3,12
@@ -704,7 +856,7 @@ class Test_Cholesky_Decomposition(TestCase):
         
 
 class Test_QR_Decomposition(TestCase):
-    def test_push_forward(self):
+    def test_pushforward(self):
         (D,P,N) = 3,5,10
         A_data = numpy.random.rand(D,P,N,N)
 
@@ -772,7 +924,7 @@ class Test_QR_Decomposition(TestCase):
         
             assert_almost_equal(numpy.trace(numpy.dot(Ab.T,Ad)), numpy.trace(numpy.dot(Q2b.T,Q2d)) + numpy.trace(numpy.dot(R2b.T,R2d)))
 
-    def test_push_forward_rectangular_A(self):
+    def test_pushforward_rectangular_A(self):
         (D,P,M,N) = 5,3,15,3
         A_data = numpy.random.rand(D,P,M,N)
 
@@ -833,7 +985,7 @@ class Test_QR_Decomposition(TestCase):
         Q2 = Q[:,N:]
         assert_array_almost_equal(0, UTPM.dot(A.T, Q2).data)
         
-    def test_push_forward_more_cols_than_rows(self):
+    def test_pushforward_more_cols_than_rows(self):
         """
         A.shape = (3,11)
         """
@@ -1088,8 +1240,36 @@ class Test_QR_Decomposition(TestCase):
 
         
 class Test_Eigenvalue_Decomposition(TestCase):
+    
+    def test_eigh1_pushforward(self):
+        (D,P,N) = 2,1,2
+        A = UTPM(numpy.zeros((D,P,N,N)))
+        A.data[0,0] = numpy.eye(N)
+        A.data[1,0] = numpy.diag([3,4])
+        
+        L,Q,b = UTPM.eigh1(A)
+        
+        assert_array_almost_equal(UTPM.dot(Q, UTPM.dot(L,Q.T)).data, A.data, decimal = 13)
+        
+        Lbar = UTPM.diag(UTPM(numpy.zeros((D,P,N))))
+        Lbar.data[0,0] = [0.5,0.5]
+        Qbar = UTPM(numpy.random.rand(*(D,P,N,N)))
+        
+        Abar = UTPM.pb_eigh1( Lbar, Qbar, None, A, L, Q, b)
+        
+        Abar = Abar.data[0,0]
+        Adot = A.data[1,0]
+        
+        Lbar = Lbar.data[0,0]
+        Ldot = L.data[1,0]
+        
+        Qbar = Qbar.data[0,0]
+        Qdot = Q.data[1,0]
+        
+        assert_almost_equal(numpy.trace(numpy.dot(Abar.T, Adot)), numpy.trace( numpy.dot(Lbar.T, Ldot) + numpy.dot(Qbar.T, Qdot)))    
+    
 
-    def test_push_forward(self):
+    def test_pushforward(self):
         (D,P,N) = 3,2,5
         A_data = numpy.zeros((D,P,N,N))
         for d in range(D):
@@ -1107,7 +1287,7 @@ class Test_Eigenvalue_Decomposition(TestCase):
 
         assert_array_almost_equal(UTPM.dot(Q, UTPM.dot(L,Q.T)).data, A.data, decimal = 12)
 
-    def test_push_forward_repeated_eigenvalues(self):
+    def test_pushforward_repeated_eigenvalues(self):
         D,P,N = 3,1,6
         A = UTPM(numpy.zeros((D,P,N,N)))
         V = UTPM(numpy.random.rand(D,P,N,N))
@@ -1125,7 +1305,7 @@ class Test_Eigenvalue_Decomposition(TestCase):
 
         assert_array_almost_equal(UTPM.dot(Q, UTPM.dot(L,Q.T)).data, A.data, decimal = 12)
 
-    def test_push_forward_repeated_eigenvalues_higher_order_multiple_direction(self):
+    def test_pushforward_repeated_eigenvalues_higher_order_multiple_direction(self):
         D,P,N = 4,7,6
         A = UTPM(numpy.zeros((D,P,N,N)))
         V = UTPM(numpy.random.rand(D,P,N,N))
